@@ -1,3 +1,11 @@
+#!/usr/bin/env python
+"""client library for iterating over http Server Sent Event (SSE) streams"""
+#
+# Distributed under the terms of the MIT license.
+#
+from __future__ import unicode_literals
+
+import codecs
 import re
 import time
 import warnings
@@ -6,12 +14,22 @@ import six
 
 import requests
 
+__version__ = '0.0.26'
+
+# Technically, we should support streams that mix line endings.  This regex,
+# however, assumes that a system will provide consistent line endings.
+end_of_field = re.compile(r'\r\n\r\n|\r\r|\n\n')
+
 
 class SSEClient(object):
-    def __init__(self, url, last_id=None, retry=3000, **kwargs):
+    def __init__(self, url, last_id=None, retry=3000, session=None, chunk_size=1024, **kwargs):
         self.url = url
         self.last_id = last_id
         self.retry = retry
+        self.chunk_size = chunk_size
+
+        # Optional support for passing in a requests.Session()
+        self.session = session
 
         # Any extra kwargs will be fed into the requests.get call later.
         self.requests_kwargs = kwargs
@@ -25,42 +43,73 @@ class SSEClient(object):
         self.requests_kwargs['headers']['Accept'] = 'text/event-stream'
 
         # Keep data here as it streams in
-        self.buf = u''
+        self.buf = ''
 
         self._connect()
 
     def _connect(self):
         if self.last_id:
             self.requests_kwargs['headers']['Last-Event-ID'] = self.last_id
-        self.resp = requests.get(self.url, stream=True,
-                                 **self.requests_kwargs)
+
+        # Use session if set.  Otherwise fall back to requests module.
+        requester = self.session or requests
+        self.resp = requester.get(self.url, stream=True, **self.requests_kwargs)
+        self.resp_iterator = self.iter_content()
+        self.decoder = codecs.getincrementaldecoder(
+            self.resp.encoding)(errors='replace')
 
         # TODO: Ensure we're handling redirects.  Might also stick the 'origin'
         # attribute on Events like the Javascript spec requires.
         self.resp.raise_for_status()
 
+    def iter_content(self):
+        def generate():
+            while True:
+                if hasattr(self.resp.raw, '_fp') and \
+                        hasattr(self.resp.raw._fp, 'fp') and \
+                        hasattr(self.resp.raw._fp.fp, 'read1'):
+                    chunk = self.resp.raw._fp.fp.read1(self.chunk_size)
+                else:
+                    # _fp is not available, this means that we cannot use short
+                    # reads and this will block until the full chunk size is
+                    # actually read
+                    chunk = self.resp.raw.read(self.chunk_size)
+                if not chunk:
+                    break
+                yield chunk
+
+        return generate()
+
+    def _event_complete(self):
+        return re.search(end_of_field, self.buf) is not None
+
     def __iter__(self):
         return self
 
     def __next__(self):
-        # TODO: additionally support CR and CRLF-style newlines.
-        while '\n\n' not in self.buf:
+        while not self._event_complete():
             try:
-                nextchar = next(self.resp.iter_content(decode_unicode=True))
-                self.buf += nextchar
-            except (StopIteration, requests.RequestException):
+                next_chunk = next(self.resp_iterator)
+                if not next_chunk:
+                    raise EOFError()
+                self.buf += self.decoder.decode(next_chunk)
+
+            except (StopIteration, requests.RequestException, EOFError, six.moves.http_client.IncompleteRead) as e:
+                print(e)
                 time.sleep(self.retry / 1000.0)
                 self._connect()
 
                 # The SSE spec only supports resuming from a whole message, so
                 # if we have half a message we should throw it out.
-                head, sep, tail = self.buf.rpartition('\n\n')
+                head, sep, tail = self.buf.rpartition('\n')
                 self.buf = head + sep
                 continue
 
-        head, sep, tail = self.buf.partition('\n\n')
-        self.buf = tail
-        msg = Event.parse(head)
+        # Split the complete event (up to the end_of_field) into event_string,
+        # and retain anything after the current complete event in self.buf
+        # for next time.
+        (event_string, self.buf) = re.split(end_of_field, self.buf, maxsplit=1)
+        msg = Event.parse(event_string)
 
         # If the server requests a specific retry delay, we need to honor it.
         if msg.retry:
@@ -109,18 +158,18 @@ class Event(object):
         and return a Event object.
         """
         msg = cls()
-        for line in raw.split('\n'):
+        for line in raw.splitlines():
             m = cls.sse_line_pattern.match(line)
             if m is None:
                 # Malformed line.  Discard but warn.
                 warnings.warn('Invalid SSE line: "%s"' % line, SyntaxWarning)
                 continue
 
-            name = m.groupdict()['name']
-            value = m.groupdict()['value']
+            name = m.group('name')
             if name == '':
                 # line began with a ":", so is a comment.  Ignore
                 continue
+            value = m.group('value')
 
             if name == 'data':
                 # If we already have some data, then join to it with a newline.
